"""Top level functions related to visualization of distributions."""

from arviz_stats.utils import _apply_multi_input_function
from arviz_stats.validate import validate_ci_prob


def hdi(
    data,
    prob=None,
    dim=None,
    group="posterior",
    var_names=None,
    filter_vars=None,
    coords=None,
    method="nearest",
    circular=False,
    max_modes=10,
    skipna=False,
    **kwargs,
):
    r"""Compute the highest density interval (HDI) given a probability.

    The HDI is the shortest interval that contains the specified probability mass.

    Parameters
    ----------
    data : array-like, DataArray, Dataset, DataTree, DataArrayGroupBy, DatasetGroupBy, or idata-like
        Input data. It will have different pre-processing applied to it depending on its type:

        - array-like: call array layer within ``arviz-stats``.
        - xarray object: apply dimension aware function to all relevant subsets
        - others: passed to :func:`arviz_base.convert_to_dataset` then treated as
          :class:`xarray.Dataset`. This option is discouraged due to needing this conversion
          which is completely automated and will be needed again in future executions or
          similar functions.

          It is recommended to first perform the conversion manually and then call
          ``arviz_stats.hdi``. This allows controlling the conversion step and inspecting
          its results.
    prob : float, optional
        Probability for the credible interval. Defaults to ``rcParams["stats.ci_prob"]``
    dim : sequence of hashable, optional
        Dimensions to be reduced when computing the HDI. Default ``rcParams["data.sample_dims"]``.
    group : hashable, default "posterior"
        Group on which to compute the HDI.
    var_names : str or list of str, optional
        Names of the variables for which the HDI should be computed.
    filter_vars : {None, "like", "regex"}, default None
    coords : dict, optional
        Dictionary of dimension/index names to coordinate values defining a subset
        of the data for which to perform the computation.
    method : str, default "nearest"
        Valid options are "nearest", "multimodal" or "multimodal_sample"
    circular : bool, default False
        Whether to compute the HDI taking into account that `data` represent circular variables
        (in the range [-np.pi, np.pi]) or not. Defaults to False (i.e non-circular variables).
    max_modes : int, default 10
        Maximum number of modes to consider when computing the HDI using the multimodal method.
    skipna : bool, default False
        If true ignores nan values when computing the HDI.
    **kwargs : any, optional
        Forwarded to the array or dataarray interface for HDI.

    Returns
    -------
    ndarray, DataArray, Dataset, DataTree
        Requested HDI of the provided input. It will have a ``ci_bound`` dimension
        with coordinate values "lower" and "upper" indicating the two extremes of
        the credible interval. In addition when using a multimodal `method` a
        ``mode`` dimension is also added.

    See Also
    --------
    arviz_stats.eti : Calculate the equal tail interval (ETI).
    arviz_stats.summary : Calculate summary statistics and diagnostics.

    Examples
    --------
    Calculate the HDI of a Normal random variable:

    .. ipython::

        In [1]: import arviz_stats as azs
           ...: import numpy as np
           ...: data = np.random.default_rng().normal(size=2000)
           ...: azs.hdi(data, 0.68)

    Calculate the HDI for specific variables:

    .. ipython::

        In [1]: import arviz_base as azb
           ...: dt = azb.load_arviz_data("centered_eight")
           ...: azs.hdi(dt, var_names=["mu", "theta"])

    Calculate the HDI also over the school dimension (for variables where present):

    .. ipython::

        In [1]: azs.hdi(dt, dim=["chain","draw", "school"])
    """
    prob = validate_ci_prob(prob)
    return _apply_multi_input_function(
        "hdi",
        data,
        dim,
        "dim",
        group=group,
        var_names=var_names,
        filter_vars=filter_vars,
        coords=coords,
        prob=prob,
        method=method,
        circular=circular,
        max_modes=max_modes,
        skipna=skipna,
        **kwargs,
    )


def eti(
    data,
    prob=None,
    dim=None,
    group="posterior",
    var_names=None,
    filter_vars=None,
    coords=None,
    method="linear",
    skipna=False,
    **kwargs,
):
    r"""Compute the  equal tail interval (ETI) given a probability.

    The ETI is constructed by dividing the remaining probability (e.g., 6% for a 94% interval)
    equally between the two tails of a distribution. Other names for ETI are central interval and
    quantile-based interval.

    Parameters
    ----------
    data : array-like, DataArray, Dataset, DataTree, DataArrayGroupBy, DatasetGroupBy, or idata-like
        Input data. It will have different pre-processing applied to it depending on its type:

        - array-like: call array layer within ``arviz-stats``.
        - xarray object: apply dimension aware function to all relevant subsets
        - others: passed to :func:`arviz_base.convert_to_dataset` then treated as
          :class:`xarray.Dataset`. This option is discouraged due to needing this conversion
          which is completely automated and will be needed again in future executions or
          similar functions.

          It is recommended to first perform the conversion manually and then call
          ``arviz_stats.eti``. This allows controlling the conversion step and inspecting
          its results.
    prob : float, optional
        Probability for the credible interval. Defaults to ``rcParams["stats.ci_prob"]``
    dim : sequence of hashable, optional
        Dimensions to be reduced when computing the HDI. Default ``rcParams["data.sample_dims"]``.
    group : hashable, default "posterior"
        Group on which to compute the ETI.
    var_names : str or list of str, optional
        Names of the variables for which the ETI should be computed.
    filter_vars : {None, "like", "regex"}, default None
    coords : dict, optional
        Dictionary of dimension/index names to coordinate values defining a subset
        of the data for which to perform the computation.
    method : str, default "linear"
        For other options see :func:`numpy.quantile`.
    skipna : bool, default False
        If true ignores nan values when computing the ETI.
    **kwargs : any, optional
        Forwarded to the array or dataarray interface for ETI.

    Returns
    -------
    ndarray, DataArray, Dataset, DataTree
        Requested ETI of the provided input. It will have a ``ci_bound`` dimension
        with coordinate values "lower" and "upper" indicating the two extremes of
        the credible interval.

    See Also
    --------
    arviz_stats.hdi : Calculate the highest density interval (HDI).
    arviz_stats.summary : Calculate summary statistics and diagnostics.

    Examples
    --------
    Calculate the ETI of a Normal random variable:

    .. ipython::

        In [1]: import arviz_stats as azs
           ...: import numpy as np
           ...: data = np.random.default_rng().normal(size=2000)
           ...: azs.eti(data, 0.68)

    Calculate the ETI for specific variables:

    .. ipython::

        In [1]: import arviz_base as azb
           ...: dt = azb.load_arviz_data("centered_eight")
           ...: azs.eti(dt, var_names=["mu", "theta"])

    Calculate the ETI also over the school dimension (for variables where present):

    .. ipython::

        In [1]: azs.eti(dt, dim=["chain","draw", "school"])
    """
    prob = validate_ci_prob(prob)
    return _apply_multi_input_function(
        "eti",
        data,
        dim,
        "dim",
        group=group,
        var_names=var_names,
        filter_vars=filter_vars,
        coords=coords,
        prob=prob,
        method=method,
        skipna=skipna,
        **kwargs,
    )


def ecdf(
    data,
    dim=None,
    group="posterior",
    var_names=None,
    filter_vars=None,
    coords=None,
    pit=False,
    **kwargs,
):
    r"""Compute the marginal empirical cumulative density functions (ECDF).

    See the EABM chapter on `Visualization of Random Variables with ArviZ <https://arviz-devs.github.io/EABM/Chapters/Distributions.html#distributions-in-arviz>`_
    for more details.

    Parameters
    ----------
    data : array-like, DataArray, Dataset, DataTree, DataArrayGroupBy, DatasetGroupBy, or idata-like
        Input data. It will have different pre-processing applied to it depending on its type:

        - array-like: call array layer within ``arviz-stats``.
        - xarray object: apply dimension aware function to all relevant subsets
        - others: passed to :func:`arviz_base.convert_to_dataset` then treated as
          :class:`xarray.Dataset`. This option is discouraged due to needing this conversion
          which is completely automated and will be needed again in future executions or
          similar functions.

          It is recommended to first perform the conversion manually and then call
          ``arviz_stats.ecdf``. This allows controlling the conversion step and inspecting
          its results.
    dim : sequence of hashable, optional
        Dimensions to be reduced when computing the ECDF. Default ``rcParams["data.sample_dims"]``.
    group : hashable, default "posterior"
        Group on which to compute the ECDF
    var_names : str or list of str, optional
        Names of the variables for which the ECDF should be computed.
    filter_vars : {None, "like", "regex"}, default None
    coords : dict, optional
        Dictionary of dimension/index names to coordinate values defining a subset
        of the data for which to perform the computation.
    pit : bool, default False
    **kwargs : any, optional
        Forwarded to the array or dataarray interface for ECDF.

    Returns
    -------
    ndarray, DataArray, Dataset, DataTree
        Requested ECDF of the provided input.
        It will have a ``quantile`` dimension and a ``plot_axis`` dimension with coordinate
        values "x" and "y".

    See Also
    --------
    arviz_stats.histogram, arviz_stats.kde, arviz_stats.qds :
        Alternative visual summaries for marginal distributions
    arviz_plots.plot_dist

    Examples
    --------
    Calculate the ECDF of a Normal random variable:

    .. ipython::

        In [1]: import arviz_stats as azs
           ...: import numpy as np
           ...: data = np.random.default_rng().normal(size=2000)
           ...: # not available yet in array interface azs.ecdf(data)

    Calculate the ECDF for specific variables:

    .. ipython::

        In [1]: import arviz_base as azb
           ...: dt = azb.load_arviz_data("centered_eight")
           ...: azs.ecdf(dt.posterior.dataset, var_names=["mu", "theta"])

    Calculate the ECDF also over the school dimension (for variables where present):

    .. ipython::

        In [1]: azs.ecdf(dt.posterior.dataset, dim=["chain", "draw", "school"])
    """
    return _apply_multi_input_function(
        "ecdf",
        data,
        dim,
        "dim",
        group=group,
        var_names=var_names,
        filter_vars=filter_vars,
        coords=coords,
        pit=pit,
        **kwargs,
    )


def histogram(
    data,
    dim=None,
    group="posterior",
    var_names=None,
    filter_vars=None,
    coords=None,
    bins=None,
    range=None,  # pylint: disable=redefined-builtin
    weights=None,
    density=True,
):
    r"""Compute the batched histogram.

    See the EABM chapter on `Visualization of Random Variables with ArviZ <https://arviz-devs.github.io/EABM/Chapters/Distributions.html#distributions-in-arviz>`_
    for more details.

    Parameters
    ----------
    data : array-like, DataArray, Dataset, DataTree, DataArrayGroupBy, DatasetGroupBy, or idata-like
        Input data. It will have different pre-processing applied to it depending on its type:

        - array-like: call array layer within ``arviz-stats``.
        - xarray object: apply dimension aware function to all relevant subsets
        - others: passed to :func:`arviz_base.convert_to_dataset` then treated as
          :class:`xarray.Dataset`. This option is discouraged due to needing this conversion
          which is completely automated and will be needed again in future executions or
          similar functions.

          It is recommended to first perform the conversion manually and then call
          ``arviz_stats.histogram``. This allows controlling the conversion step and inspecting
          its results.
    dim : sequence of hashable, optional
        Dimensions to be reduced when computing the histogram.
        Default ``rcParams["data.sample_dims"]``.
    group : hashable, default "posterior"
        Group on which to compute the histogram
    var_names : str or list of str, optional
        Names of the variables for which the histogram should be computed.
    filter_vars : {None, "like", "regex"}, default None
    coords : dict, optional
        Dictionary of dimension/index names to coordinate values defining a subset
        of the data for which to perform the computation.
    bind : array-like, optional
    range : array-like, optional
    weights : array-like, optional
    density : bool, default True
    **kwargs : any, optional
        Forwarded to the array or dataarray interface for histogram.

    Returns
    -------
    ndarray, DataArray, Dataset, DataTree
        Requested histogram of the provided input.
        It will have a ``hist_dim_{var_name}`` dimension and a ``plot_axis`` dimension
        with coordinates "histogram", "left_edges" and "right_edges"

    See Also
    --------
    arviz_stats.ecdf, arviz_stats.kde, arviz_stats.qds :
        Alternative visual summaries for marginal distributions
    arviz_plots.plot_dist

    Examples
    --------
    Calculate the histogram of a Normal random variable:

    .. ipython::

        In [1]: import arviz_stats as azs
           ...: import numpy as np
           ...: data = np.random.default_rng().normal(size=2000)
           ...: azs.histogram(data)

    Calculate the histogram for specific variables:

    .. ipython::

        In [1]: import arviz_base as azb
           ...: dt = azb.load_arviz_data("centered_eight")
           ...: azs.histogram(dt, var_names=["mu", "theta"])

    Calculate the histogram also over the school dimension (for variables where present):

    .. ipython::

        In [1]: azs.histogram(dt, dim=["chain", "draw", "school"])
    """
    return _apply_multi_input_function(
        "histogram",
        data,
        dim,
        "dim",
        group=group,
        var_names=var_names,
        filter_vars=filter_vars,
        coords=coords,
        bins=bins,
        range=range,
        weights=weights,
        density=density,
    )


def kde(
    data,
    dim=None,
    group="posterior",
    var_names=None,
    filter_vars=None,
    coords=None,
    circular=False,
    **kwargs,
):
    r"""Compute the marginal kernel density estimates (KDE).

    See the EABM chapter on `Visualization of Random Variables with ArviZ <https://arviz-devs.github.io/EABM/Chapters/Distributions.html#distributions-in-arviz>`_
    for more details.

    Parameters
    ----------
    data : array-like, DataArray, Dataset, DataTree, DataArrayGroupBy, DatasetGroupBy, or idata-like
        Input data. It will have different pre-processing applied to it depending on its type:

        - array-like: call array layer within ``arviz-stats``.
        - xarray object: apply dimension aware function to all relevant subsets
        - others: passed to :func:`arviz_base.convert_to_dataset` then treated as
          :class:`xarray.Dataset`. This option is discouraged due to needing this conversion
          which is completely automated and will be needed again in future executions or
          similar functions.

          It is recommended to first perform the conversion manually and then call
          ``arviz_stats.kde``. This allows controlling the conversion step and inspecting
          its results.
    dim : sequence of hashable, optional
        Dimensions to be reduced when computing the KDE.
        Default ``rcParams["data.sample_dims"]``.
    group : hashable, default "posterior"
        Group on which to compute the KDE
    var_names : str or list of str, optional
        Names of the variables for which the KDE should be computed.
    filter_vars : {None, "like", "regex"}, default None
    coords : dict, optional
        Dictionary of dimension/index names to coordinate values defining a subset
        of the data for which to perform the computation.
    circular : bool, default False
    **kwargs : any, optional
        Forwarded to the array or dataarray interface for KDE.

    Returns
    -------
    ndarray, DataArray, Dataset, DataTree
        Requested KDE of the provided input.
        The xarray objects will have a ``kde_dim`` dimension and a ``plot_axis`` dimension
        with coordinates "x", and "y".

    See Also
    --------
    arviz_stats.ecdf, arviz_stats.histogram, arviz_stats.qds:
        Alternative visual summaries for marginal distributions
    arviz_plots.plot_dist

    Examples
    --------
    Calculate the KDE of a Normal random variable:

    .. ipython::

        In [1]: import arviz_stats as azs
           ...: import numpy as np
           ...: data = np.random.default_rng().normal(size=2000)
           ...: azs.kde(data)

    Calculate the KDE for specific variables:

    .. ipython::

        In [1]: import arviz_base as azb
           ...: dt = azb.load_arviz_data("centered_eight")
           ...: azs.kde(dt, var_names=["mu", "theta"])

    Calculate the KDE also over the school dimension (for variables where present):

    .. ipython::

        In [1]: azs.kde(dt, dim=["chain", "draw", "school"])
    """
    return _apply_multi_input_function(
        "kde",
        data,
        dim,
        "dim",
        group=group,
        var_names=var_names,
        filter_vars=filter_vars,
        coords=coords,
        circular=circular,
        **kwargs,
    )


def qds(
    data,
    dim=None,
    group="posterior",
    var_names=None,
    filter_vars=None,
    coords=None,
    nquantiles=100,
    binwidth=None,
    dotsize=1,
    stackratio=1,
    **kwargs,
):
    r"""Compute the marginal quantile dots.

    For details see [1]_ and check the EABM chapter on `Visualization of Random Variables with
    ArviZ <https://arviz-devs.github.io/EABM/Chapters/Distributions.html#distributions-in-arviz>`_.

    Parameters
    ----------
    data : array-like, DataArray, Dataset, DataTree, DataArrayGroupBy, DatasetGroupBy, or idata-like
        Input data. It will have different pre-processing applied to it depending on its type:

        - array-like: call array layer within ``arviz-stats``.
        - xarray object: apply dimension aware function to all relevant subsets
        - others: passed to :func:`arviz_base.convert_to_dataset` then treated as
          :class:`xarray.Dataset`. This option is discouraged due to needing this conversion
          which is completely automated and will be needed again in future executions or
          similar functions.

          It is recommended to first perform the conversion manually and then call
          ``arviz_stats.kde``. This allows controlling the conversion step and inspecting
          its results.
    dim : sequence of hashable, optional
        Dimensions to be reduced when computing the quantile dots
        Default ``rcParams["data.sample_dims"]``.
    group : hashable, default "posterior"
        Group on which to compute the quantile dots
    var_names : str or list of str, optional
        Names of the variables for which the quantile dots should be computed.
    filter_vars : {None, "like", "regex"}, default None
    coords : dict, optional
        Dictionary of dimension/index names to coordinate values defining a subset
        of the data for which to perform the computation.
    binwidth : float, optional
        Width of the bin for the dots.
    dotsize : float, default 1
        The size of the dots relative to the bin width. The default makes dots be just about as
        wide as the bin width.
    stackratio : float, default 1
        The distance between the center of the dots in the same stack relative to the bin height.
        The default makes dots in the same stack just touch each other.
    **kwargs : any, optional
        Forwarded to the array or dataarray interface for quantile dots.

    Returns
    -------
    ndarray, DataArray, Dataset, DataTree
        Requested QDs of the provided input.
        The xarray objects will have a ``qds_dim`` dimension and a ``plot_axis`` dimension
        with coordinates "x", and "y".

    See Also
    --------
    arviz_stats.ecdf, arviz_stats.histogram, arviz_stats.kde:
        Alternative visual summaries for marginal distributions
    arviz_plots.plot_dist

    References
    ----------
    .. [1] Kay M, Kola T, Hullman JR, and Munson SA. *When (ish) is My Bus?:
       User-centered Visualizations of Uncertainty in Everyday, Mobile Predictive
       Systems.* In Proceedings of the 2016 CHI Conference Association for Computing
       Machinery. 2016. https://doi.org/10.1145/2858036.2858558

    Examples
    --------
    Calculate the QDs of a Normal random variable:

    .. ipython::

        In [1]: import arviz_stats as azs
           ...: import numpy as np
           ...: data = np.random.default_rng().normal(size=2000)
           ...: azs.qds(data)

    Calculate the QDs for specific variables:

    .. ipython::

        In [1]: import arviz_base as azb
           ...: dt = azb.load_arviz_data("centered_eight")
           ...: azs.qds(dt, var_names=["mu", "theta"])

    Calculate the QDs also over the school dimension (for variables where present):

    .. ipython::

        In [1]: azs.qds(dt, dim=["chain", "draw", "school"])
    """
    return _apply_multi_input_function(
        "qds",
        data,
        dim,
        "dim",
        group=group,
        var_names=var_names,
        filter_vars=filter_vars,
        coords=coords,
        nquantiles=nquantiles,
        binwidth=binwidth,
        dotsize=dotsize,
        stackratio=stackratio,
        **kwargs,
    )



"""Helper functions for K-fold cross-validation."""

from collections import namedtuple

import numpy as np
import xarray as xr
from arviz_base import convert_to_datatree
from xarray_einstats.stats import logsumexp

from arviz_stats.loo.wrapper import SamplingWrapper
from arviz_stats.utils import get_log_likelihood

__all__ = [
    "_prepare_kfold_inputs",
    "_compute_kfold_results",
    "_kfold_split_random",
    "_kfold_split_stratified",
    "_kfold_split_grouped",
    "_extract_fold_data",
    "_get_fold_indices",
    "_combine_fold_elpds",
    "_validate_k_value",
    "_validate_array_length",
    "_validate_fold_parameters",
]

FoldData = namedtuple("FoldData", ["train_indices", "test_indices", "fold_id"])

KfoldResults = namedtuple("KfoldResults", ["elpds", "ps", "fold_fits", "elpd_i", "p_kfold_i"])

KfoldInputs = namedtuple(
    "KfoldInputs",
    [
        "log_likelihood",
        "var_name",
        "sample_dims",
        "obs_dims",
        "n_data_points",
        "n_samples",
        "folds",
        "k",
    ],
)


def _prepare_kfold_inputs(data, var_name, wrapper, k, folds, stratify_by, group_by):
    """Prepare inputs for k-fold cross-validation."""
    data = convert_to_datatree(data)

    if not isinstance(wrapper, SamplingWrapper):
        raise TypeError("wrapper must be an instance of SamplingWrapper")

    required_methods = ["sel_observations", "sample", "get_inference_data", "log_likelihood__i"]
    not_implemented = wrapper.check_implemented_methods(required_methods)

    if not_implemented:
        raise ValueError(
            f"The following methods must be implemented in the SamplingWrapper: {not_implemented}"
        )

    log_likelihood = get_log_likelihood(data, var_name)
    var_name = log_likelihood.name if var_name is None else var_name
    sample_dims = ["chain", "draw"]

    obs_dims = [dim for dim in log_likelihood.dims if dim not in sample_dims]
    n_data_points = int(np.prod([log_likelihood.sizes[dim] for dim in obs_dims]))
    n_samples = int(np.prod([log_likelihood.sizes[dim] for dim in sample_dims]))

    _validate_fold_parameters(folds, stratify_by, group_by)

    if folds is not None:
        folds = _validate_array_length(folds, n_data_points, "folds")
        k = len(np.unique(folds))
    elif stratify_by is not None:
        stratify_by = _validate_array_length(stratify_by, n_data_points, "stratify_by")
        folds = _kfold_split_stratified(k=k, x=stratify_by)
    elif group_by is not None:
        group_by = _validate_array_length(group_by, n_data_points, "group_by")
        folds = _kfold_split_grouped(k=k, x=group_by)
    else:
        folds = _kfold_split_random(k=k, n=n_data_points)

    return KfoldInputs(
        log_likelihood=log_likelihood,
        var_name=var_name,
        sample_dims=sample_dims,
        obs_dims=obs_dims,
        n_data_points=n_data_points,
        n_samples=n_samples,
        folds=folds,
        k=k,
    )


def _compute_kfold_results(kfold_inputs, wrapper, save_fits):
    """Compute k-fold cross-validation results."""
    ll_full = kfold_inputs.log_likelihood
    lpds_full = logsumexp(ll_full, dims=kfold_inputs.sample_dims, b=1 / kfold_inputs.n_samples)

    fold_indices = _get_fold_indices(kfold_inputs.folds, kfold_inputs.k)

    elpd_unordered = []
    fold_results = {} if save_fits else None

    for fold_num in range(1, kfold_inputs.k + 1):
        test_idx = fold_indices[fold_num]["test_indices"]
        train_data, test_data = wrapper.sel_observations(test_idx)
        fitted_model = wrapper.sample(train_data)
        idata_k = wrapper.get_inference_data(fitted_model)
        log_lik_k = wrapper.log_likelihood__i(test_data, idata_k)

        if "chain" in log_lik_k.dims and "draw" in log_lik_k.dims:
            sample_dims_k = ["chain", "draw"]
        else:
            sample_dims_k = [dim for dim in log_lik_k.dims if dim not in kfold_inputs.obs_dims]

        n_samples_k = int(np.prod([log_lik_k.sizes[dim] for dim in sample_dims_k]))
        elpd_k = logsumexp(log_lik_k, dims=sample_dims_k, b=1 / n_samples_k)
        elpd_values = elpd_k.values if hasattr(elpd_k, "values") else elpd_k

        if np.isscalar(elpd_values):
            elpd_unordered.append((test_idx[0], float(elpd_values)))
        else:
            for i, idx in enumerate(test_idx):
                elpd_unordered.append((idx, float(elpd_values[i])))

        if save_fits:
            fold_results[fold_num] = {"fit": idata_k, "test_indices": test_idx}

    elpd_unordered.sort(key=lambda x: x[0])
    elpds = np.array([x[1] for x in elpd_unordered])

    lpds_ordered = lpds_full.values.flatten()
    ps = lpds_ordered - elpds

    obs_dim = kfold_inputs.obs_dims[-1] if kfold_inputs.obs_dims else "obs"

    elpd_i = xr.DataArray(
        elpds,
        dims=[obs_dim],
        coords={obs_dim: kfold_inputs.log_likelihood.coords[obs_dim]},
    )

    p_kfold_i = xr.DataArray(
        ps,
        dims=[obs_dim],
        coords={obs_dim: kfold_inputs.log_likelihood.coords[obs_dim]},
    )

    return KfoldResults(
        elpds=elpds, ps=ps, fold_fits=fold_results, elpd_i=elpd_i, p_kfold_i=p_kfold_i
    )


def _kfold_split_random(k=10, n=None):
    """Split the data into K groups of equal size (or roughly equal size)."""
    if n is None:
        raise ValueError("n must be provided")
    if not isinstance(n, int | np.integer) or np.isnan(n):
        raise ValueError("n must be an integer")
    n = int(n)
    k = _validate_k_value(k, n)

    fold_size = n // k
    remainder = n % k

    folds = np.zeros(n, dtype=int)
    start = 0
    for i in range(k):
        end = start + fold_size + (1 if i < remainder else 0)
        folds[start:end] = i + 1
        start = end

    rng = np.random.default_rng()
    perm = rng.permutation(n)
    return folds[perm]


def _kfold_split_stratified(k=10, x=None):
    """Split observations into k groups ensuring relative category frequencies are preserved."""
    if x is None:
        raise ValueError("x must be provided")
    x = np.asarray(x)
    k = _validate_k_value(k, len(x))

    unique_vals = np.unique(x)
    x_int = np.zeros(len(x), dtype=int)
    for i, val in enumerate(unique_vals):
        x_int[x == val] = i

    n_lev = len(unique_vals)
    n = len(x)
    xids = []

    rng = np.random.default_rng()
    for level in range(n_lev):
        idx = np.where(x_int == level)[0]
        if len(idx) > 1:
            xids.extend(rng.permutation(idx))
        else:
            xids.extend(idx)

    xids = np.array(xids)
    bins = np.zeros(n, dtype=int)
    bins[xids] = np.tile(np.arange(1, k + 1), int(np.ceil(n / k)))[:n]

    return bins


def _kfold_split_grouped(k=10, x=None):
    """Split observations ensuring all observations from the same group stay together."""
    if x is None:
        raise ValueError("x must be provided")
    x = np.asarray(x)
    k = _validate_k_value(k, len(x))

    unique_vals = np.unique(x)
    x_int = np.zeros(len(x), dtype=int)
    for i, val in enumerate(unique_vals):
        x_int[x == val] = i + 1

    n_levels = len(unique_vals)

    if n_levels < k:
        raise ValueError("k must not be bigger than the number of levels/groups in x")

    if n_levels == k:
        return x_int

    s1 = int(np.ceil(n_levels / k))
    n_s2 = s1 * k - n_levels
    n_s1 = k - n_s2

    rng = np.random.default_rng()
    perm = rng.permutation(n_levels) + 1

    breaks = []
    if n_s1 > 0:
        breaks.extend(np.arange(s1 + 0.5, s1 * n_s1 + 0.5, s1))
    if n_s2 > 0:
        start = breaks[-1] if breaks else 0.5
        breaks.extend(np.arange(start + s1 - 1, start + (s1 - 1) * n_s2, s1 - 1))

    breaks = np.array(breaks)
    groups = np.searchsorted(breaks, perm, side="right") + 1

    bins = np.zeros(len(x), dtype=int)
    for j in range(1, n_levels + 1):
        bins[x_int == j] = groups[j - 1]

    return bins


def _extract_fold_data(data, fold_indices, train=True):
    """Extract data for a specific fold."""
    if train:
        mask = np.ones(data.shape[-1], dtype=bool)
        mask[fold_indices] = False
        return data.isel({data.dims[-1]: mask})
    return data.isel({data.dims[-1]: fold_indices})


def _get_fold_indices(fold_assignments, k):
    """Get test indices for each fold."""
    results = {}

    for i in range(1, k + 1):
        test_indices = np.where(fold_assignments == i)[0]
        results[i] = {"test_indices": test_indices, "n_test": len(test_indices)}

    return results


def _combine_fold_elpds(fold_elpds, n_data_points):
    """Combine ELPD values from all folds into final estimates."""
    elpds = np.concatenate(fold_elpds)
    elpd_kfold = np.sum(elpds)
    se_elpd_kfold = np.sqrt(n_data_points * np.var(elpds))

    return {"elpd_kfold": elpd_kfold, "se_elpd_kfold": se_elpd_kfold, "pointwise": elpds}


def _validate_k_value(k, n, param_name="k"):
    """Validate k parameter for k-fold splitting."""
    if not isinstance(k, int | np.integer) or np.isnan(k):
        raise ValueError(f"{param_name} must be an integer")
    k = int(k)
    if k <= 1:
        raise ValueError(f"{param_name} must be greater than 1")
    if k > n:
        raise ValueError(f"{param_name} must not be greater than n ({n})")
    return k


def _validate_array_length(array, expected_length, param_name):
    """Validate array length matches expected number of observations."""
    if array is None:
        raise ValueError(f"{param_name} must be provided")

    if isinstance(array, xr.DataArray):
        array_length = array.size
        array_values = array.values.flatten()
    else:
        array = np.asarray(array)
        array_length = len(array)
        array_values = array

    if array_length != expected_length:
        raise ValueError(
            f"Length of {param_name} ({array_length}) must match number of "
            f"observations ({expected_length})"
        )
    return array_values


def _validate_fold_parameters(folds, stratify_by, group_by):
    """Validate fold parameter combinations."""
    if folds is not None:
        if stratify_by is not None or group_by is not None:
            raise ValueError(
                "Cannot use stratify_by or group_by when folds are explicitly provided"
            )
    elif stratify_by is not None and group_by is not None:
        raise ValueError("Cannot use both stratify_by and group_by")



"""Pareto-smoothed importance sampling LOO (PSIS-LOO-CV) and K-fold cross-validation functions."""

from arviz_stats.loo.loo import loo, loo_i
from arviz_stats.loo.loo_approximate_posterior import loo_approximate_posterior
from arviz_stats.loo.loo_expectations import loo_expectations, loo_metrics
from arviz_stats.loo.loo_score import loo_score
from arviz_stats.loo.loo_pit import loo_pit
from arviz_stats.loo.loo_subsample import loo_subsample, update_subsample
from arviz_stats.loo.loo_moment_match import loo_moment_match
from arviz_stats.loo.reloo import reloo
from arviz_stats.loo.wrapper import SamplingWrapper
from arviz_stats.loo.compare import compare, _calculate_ics
from arviz_stats.loo.loo_kfold import loo_kfold

__all__ = [
    "loo",
    "loo_i",
    "loo_approximate_posterior",
    "loo_expectations",
    "loo_metrics",
    "loo_score",
    "loo_pit",
    "loo_subsample",
    "update_subsample",
    "loo_moment_match",
    "reloo",
    "SamplingWrapper",
    "compare",
    "loo_kfold",
]



"""Compute PSIS-LOO-CV for approximate posteriors."""

from arviz_base import rcParams

from arviz_stats.loo.helper_loo import (  # pylint: disable=cyclic-import
    _check_log_density,
    _compute_loo_results,
    _prepare_loo_inputs,
)


def loo_approximate_posterior(data, log_p, log_q, pointwise=None, var_name=None, log_jacobian=None):
    r"""Compute PSIS-LOO-CV for approximate posteriors.

    Estimates the expected log pointwise predictive density (elpd) using Pareto-smoothed
    importance sampling leave-one-out cross-validation (PSIS-LOO-CV) for approximate
    posteriors (e.g., from variational inference). Requires log-densities of the target (log_p)
    and proposal (log_q) distributions.

    The PSIS-LOO-CV method is described in [1]_ and [2]_. The approximate posterior correction
    is computed using the method described in [3]_.

    Parameters
    ----------
    data : DataTree or InferenceData
        Input data. It should contain the log_likelihood group corresponding to samples
        drawn from the proposal distribution (q).
    log_p : ndarray or DataArray
        The (target) log-density evaluated at S samples from the target distribution (p).
        If ndarray, should be a vector of length S where S is the number of samples.
        If DataArray, should have dimensions matching the sample dimensions
        ("chain", "draw").
    log_q : ndarray or DataArray
        The (proposal) log-density evaluated at S samples from the proposal distribution (q).
        If ndarray, should be a vector of length S where S is the number of samples.
        If DataArray, should have dimensions matching the sample dimensions
        ("chain", "draw").
    pointwise : bool, optional
        If True, returns pointwise values. Defaults to rcParams["stats.ic_pointwise"].
    var_name : str, optional
        The name of the variable in log_likelihood groups storing the pointwise log
        likelihood data to use for loo computation.
    log_jacobian : DataArray, optional
        Log-Jacobian adjustment for variable transformations. Required when the model was fitted
        on transformed response data :math:`z = T(y)` but you want to compute ELPD on the
        original response scale :math:`y`. The value should be :math:`\log|\frac{dz}{dy}|`
        (the log absolute value of the derivative of the transformation). Must be a DataArray
        with dimensions matching the observation dimensions.

    Returns
    -------
    ELPDData
        Object with the following attributes:

        - **kind**: "loo"
        - **elpd**: expected log pointwise predictive density
        - **se**: standard error of the elpd
        - **p**: effective number of parameters
        - **n_samples**: number of samples
        - **n_data_points**: number of data points
        - **scale**: "log"
        - **warning**: True if the estimated shape parameter of Pareto distribution is greater
          than ``good_k``.
        - **good_k**: For a sample size S, the threshold is computed as
          ``min(1 - 1/log10(S), 0.7)``
        - **elpd_i**: :class:`~xarray.DataArray` with the pointwise predictive accuracy,
          only if ``pointwise=True``
        - **pareto_k**: :class:`~xarray.DataArray` with Pareto shape values,
          only if ``pointwise=True``
        - **approx_posterior**: True (approximate posterior correction applied)

    Examples
    --------
    To calculate PSIS-LOO-CV for posterior approximations, we need to provide the log-densities
    of the target and proposal distributions. Here we use dummy log-densities. In practice, the
    log-densities would typically be computed by a posterior approximation method such as the
    Laplace approximation or automatic differentiation variational inference (ADVI):

    .. ipython::

        In [1]: import numpy as np
           ...: import xarray as xr
           ...: from arviz_stats import loo_approximate_posterior
           ...: from arviz_base import load_arviz_data, extract
           ...:
           ...: data = load_arviz_data("centered_eight")
           ...: log_lik = extract(data, group="log_likelihood", var_names="obs", combined=False)
           ...: rng = np.random.default_rng(214)
           ...:
           ...: values_p = rng.normal(loc=0, scale=1, size=(log_lik.chain.size, log_lik.draw.size))
           ...: log_p = xr.DataArray(
           ...:     values_p,
           ...:     dims=["chain", "draw"],
           ...:     coords={"chain": log_lik.chain, "draw": log_lik.draw}
           ...: )
           ...:
           ...: values_q = rng.normal(loc=-1, scale=1, size=(log_lik.chain.size, log_lik.draw.size))
           ...: log_q = xr.DataArray(
           ...:     values_q,
           ...:     dims=["chain", "draw"],
           ...:     coords={"chain": log_lik.chain, "draw": log_lik.draw}
           ...: )

    Now we can calculate pointwise PSIS-LOO-CV for posterior approximations:

    .. ipython::

        In [2]: loo_approx = loo_approximate_posterior(
           ...:     data,
           ...:     log_p=log_p,
           ...:     log_q=log_q,
           ...:     var_name="obs",
           ...:     pointwise=True
           ...: )
           ...: loo_approx

    We can also calculate the PSIS-LOO-CV for posterior approximations with subsampling
    for large datasets:

    .. ipython::

        In [3]: from arviz_stats import loo_subsample
           ...: loo_approx_subsample = loo_subsample(
           ...:     data,
           ...:     observations=4,
           ...:     var_name="obs",
           ...:     log_p=log_p,
           ...:     log_q=log_q,
           ...:     pointwise=True
           ...: )
           ...: loo_approx_subsample

    See Also
    --------
    loo : Standard PSIS-LOO-CV.
    loo_subsample : Sub-sampled PSIS-LOO-CV.
    compare : Compare models based on their ELPD.

    References
    ----------

    .. [1] Vehtari et al. *Practical Bayesian model evaluation using leave-one-out cross-validation
        and WAIC*. Statistics and Computing. 27(5) (2017) https://doi.org/10.1007/s11222-016-9696-4
        arXiv preprint https://arxiv.org/abs/1507.04544.

    .. [2] Vehtari et al. *Pareto Smoothed Importance Sampling*.
        Journal of Machine Learning Research, 25(72) (2024) https://jmlr.org/papers/v25/19-556.html
        arXiv preprint https://arxiv.org/abs/1507.02646

    .. [3] Magnusson, M., Riis Andersen, M., Jonasson, J., & Vehtari, A. *Bayesian Leave-One-Out
        Cross-Validation for Large Data.* Proceedings of the 36th International Conference on
        Machine Learning, PMLR 97:4244–4253 (2019)
        https://proceedings.mlr.press/v97/magnusson19a.html
        arXiv preprint https://arxiv.org/abs/1904.10679
    """
    loo_inputs = _prepare_loo_inputs(data, var_name)
    pointwise = rcParams["stats.ic_pointwise"] if pointwise is None else pointwise

    log_likelihood = loo_inputs.log_likelihood
    log_p = _check_log_density(
        log_p, "log_p", log_likelihood, loo_inputs.n_samples, loo_inputs.sample_dims
    )
    log_q = _check_log_density(
        log_q, "log_q", log_likelihood, loo_inputs.n_samples, loo_inputs.sample_dims
    )

    approx_correction = log_p - log_q

    # Handle underflow/overflow
    approx_correction = approx_correction - approx_correction.max()

    corrected_log_ratios = -log_likelihood.copy()
    corrected_log_ratios = corrected_log_ratios + approx_correction

    # Handle underflow/overflow
    log_ratio_max = corrected_log_ratios.max(dim=loo_inputs.sample_dims)
    corrected_log_ratios = corrected_log_ratios - log_ratio_max

    # ignore r_eff here, set to r_eff=1.0
    psis_input = -corrected_log_ratios
    log_weights, pareto_k = psis_input.azstats.psislw(r_eff=1.0, dim=loo_inputs.sample_dims)

    return _compute_loo_results(
        log_likelihood=loo_inputs.log_likelihood,
        var_name=loo_inputs.var_name,
        pointwise=pointwise,
        sample_dims=loo_inputs.sample_dims,
        n_samples=loo_inputs.n_samples,
        n_data_points=loo_inputs.n_data_points,
        log_weights=log_weights,
        pareto_k=pareto_k,
        approx_posterior=True,
        log_jacobian=log_jacobian,
    )



"""K-fold cross-validation."""

import numpy as np
from arviz_base import rcParams

from arviz_stats.loo.helper_loo_kfold import (
    _combine_fold_elpds,
    _compute_kfold_results,
    _prepare_kfold_inputs,
)
from arviz_stats.utils import ELPDData


def loo_kfold(
    data,
    wrapper,
    pointwise=None,
    var_name=None,
    k=10,
    folds=None,
    stratify_by=None,
    group_by=None,
    save_fits=False,
):
    """Perform exact K-fold cross-validation.

    K-fold cross-validation evaluates model predictive accuracy by partitioning the data
    into K complementary subsets (folds), then iteratively refitting the model K times,
    each time holding out one fold as a test set and training on the remaining K-1 folds.

    This method provides an unbiased estimate of model performance by ensuring each
    observation is used exactly once for testing. Unlike PSIS-LOO-CV (Pareto-smoothed
    importance sampling leave-one-out cross-validation), which approximates cross-validation
    efficiently, K-fold requires actual model refitting but yields exact results.

    Parameters
    ----------
    data : DataTree or InferenceData
        Input data containing the posterior and log_likelihood groups from the full model fit.
    wrapper : SamplingWrapper
        An instance of SamplingWrapper class handling model refitting. The wrapper must
        implement the following methods: sel_observations, sample, get_inference_data,
        and log_likelihood__i.
    pointwise : bool, optional
        If True, return pointwise estimates. Defaults to ``rcParams["stats.ic_pointwise"]``.
    var_name : str, optional
        The name of the variable in log_likelihood group storing the pointwise log
        likelihood data to use for computation.
    k : int, default=10
        The number of folds for cross-validation. The data will be partitioned into k subsets
        of equal (or approximately equal) size.
    folds : array or DataArray, optional
        Manual fold assignments (1 to k) for each observation. For example, [1,1,2,2,3,3,4,4]
        assigns first two obs to fold 1, next two to fold 2, etc. If not provided, creates k
        random folds of equal size. Cannot be used together with `stratify_by` and `group_by`.
    stratify_by : array or DataArray, optional
        Maintains class proportions across folds. For example, [0,0,1,1,0,0,1,1] ensures each
        fold has 50% class 0 and 50% class 1. Cannot be used together with `folds` and `group_by`.
    group_by : array or DataArray, optional
        Grouping variable to keep related observations together in the same fold. For example,
        [1,1,2,2,3,3] keeps all obs from group 1 in one fold, group 2 in another, etc. Useful
        for repeated measures or clustered data. Cannot be used together with `folds` and
        `stratify_by`.
    save_fits : bool, default=False
        If True, store the fitted models and fold indices in the returned object.

    Returns
    -------
    ELPDData
        Object with the following attributes:

        - **kind**: "loo_kfold"
        - **elpd**: expected log pointwise predictive density
        - **se**: standard error of the elpd
        - **p**: effective number of parameters
        - **n_samples**: number of samples per fold
        - **n_data_points**: number of data points
        - **scale**: "log"
        - **warning**: False (not applicable for :math:`k`-fold)
        - **good_k**: None (not applicable for :math:`k`-fold)
        - **elpd_i**: :class:`~xarray.DataArray` with pointwise predictive accuracy,
          only if ``pointwise=True``
        - **pareto_k**: None (not applicable for :math:`k`-fold)
        - **n_folds**: number of folds (:math:`k`)
        - **p_kfold_i**: :class:`~xarray.DataArray` with pointwise effective number of parameters,
          only if ``pointwise=True``
        - **fold_fits**: Dictionary containing fitted models for each fold,
          only if ``save_fits=True``

    Examples
    --------
    Unlike PSIS-LOO (which approximates LOO-CV), k-fold cross-validation refits
    the model k times. So we need to tell ``loo_kfold`` how to refit the model.

    This is done by creating an instance of the ``SamplingWrapper`` class that
    implements four key methods: ``sel_observations``, ``sample``, ``get_inference_data``,
    and ``log_likelihood__i``.

    .. ipython::

        In [1]: import numpy as np
           ...: import xarray as xr
           ...: from scipy import stats
           ...: from arviz_base import load_arviz_data, from_dict
           ...: from arviz_stats import loo_kfold
           ...: from arviz_stats.loo import SamplingWrapper
           ...:
           ...: class CenteredEightWrapper(SamplingWrapper):
           ...:     def __init__(self, idata):
           ...:         super().__init__(model=None, idata_orig=idata)
           ...:         self.y_obs = idata.observed_data["obs"].values
           ...:         self.sigma = np.array([15, 10, 16, 11, 9, 11, 10, 18])
           ...:
           ...:     def sel_observations(self, idx):
           ...:         all_idx = np.arange(len(self.y_obs))
           ...:         train_idx = np.setdiff1d(all_idx, idx)
           ...:
           ...:         train_data = {
           ...:             "y": self.y_obs[train_idx],
           ...:             "sigma": self.sigma[train_idx],
           ...:             "indices": train_idx
           ...:         }
           ...:         test_data = {
           ...:             "y": self.y_obs[idx],
           ...:             "sigma": self.sigma[idx],
           ...:             "indices": idx
           ...:         }
           ...:         return train_data, test_data
           ...:
           ...:     def sample(self, modified_observed_data):
           ...:         # (Simplified version where we normally would use the actual sampler)
           ...:         train_y = modified_observed_data["y"]
           ...:         n = 1000
           ...:         mu = np.random.normal(train_y.mean(), 5, n)
           ...:         tau = np.abs(np.random.normal(10, 2, n))
           ...:         return {"mu": mu, "tau": tau}
           ...:
           ...:     def get_inference_data(self, fitted_model):
           ...:         posterior = {
           ...:             "mu": fitted_model["mu"].reshape(1, -1),
           ...:             "tau": fitted_model["tau"].reshape(1, -1)
           ...:         }
           ...:         return from_dict({"posterior": posterior})
           ...:
           ...:     def log_likelihood__i(self, excluded_obs, idata__i):
           ...:         test_y = excluded_obs["y"]
           ...:         test_sigma = excluded_obs["sigma"]
           ...:         mu = idata__i.posterior["mu"].values.flatten()
           ...:         tau = idata__i.posterior["tau"].values.flatten()
           ...:
           ...:         var_total = tau[:, np.newaxis] ** 2 + test_sigma**2
           ...:         log_lik = stats.norm.logpdf(
           ...:             test_y, loc=mu[:, np.newaxis], scale=np.sqrt(var_total)
           ...:         )
           ...:
           ...:         dims = ["chain", "school", "draw"]
           ...:         coords = {"school": excluded_obs["indices"]}
           ...:         return xr.DataArray(
           ...:             log_lik.T[np.newaxis, :, :], dims=dims, coords=coords
           ...:         )

    Now let's run k-fold cross-validation. With k=4, we'll refit the model 4 times,
    each time leaving out 2 schools for testing:

    .. ipython::

        In [2]: data = load_arviz_data("centered_eight")
           ...: wrapper = CenteredEightWrapper(data)
           ...: kfold_results = loo_kfold(data, wrapper, k=4, pointwise=True)
           ...: kfold_results

    Sometimes we want more control over how the data is split. For instance,
    if you have imbalanced groups, stratified k-fold ensures each fold has
    a similar distribution:

    .. ipython::

        In [3]: strata = (data.observed_data["obs"] > 5).astype(int)
           ...: kfold_strat = loo_kfold(data, wrapper, k=4, stratify_by=strata)
           ...: kfold_strat

    Moreover, sometimes we want to group observations together. For instance,
    if we have repeated measurements from the same subject, we can group by subject:

    .. ipython::

        In [4]: groups = xr.DataArray([1, 1, 2, 2, 3, 3, 4, 4], dims="school")
           ...: kfold_group = loo_kfold(data, wrapper, k=4, group_by=groups)
           ...: kfold_group

    Notes
    -----
    When K equals the number of observations, this becomes exact leave-one-out
    cross-validation. Note that :func:`arviz_stats.loo` provides a much more efficient
    approximation for that case and is recommended.

    See Also
    --------
    loo : Pareto-smoothed importance sampling LOO-CV
    SamplingWrapper : Base class for implementing sampling wrappers

    References
    ----------

    .. [1] Vehtari et al. *Practical Bayesian model evaluation using leave-one-out cross-validation
        and WAIC*. Statistics and Computing. 27(5) (2017) https://doi.org/10.1007/s11222-016-9696-4
        arXiv preprint https://arxiv.org/abs/1507.04544.

    .. [2] Vehtari et al. *Pareto Smoothed Importance Sampling*.
        Journal of Machine Learning Research, 25(72) (2024) https://jmlr.org/papers/v25/19-556.html
        arXiv preprint https://arxiv.org/abs/1507.02646
    """
    pointwise = rcParams["stats.ic_pointwise"] if pointwise is None else pointwise

    kfold_inputs = _prepare_kfold_inputs(data, var_name, wrapper, k, folds, stratify_by, group_by)
    kfold_results = _compute_kfold_results(kfold_inputs, wrapper, save_fits)

    combined_results = _combine_fold_elpds([kfold_results.elpds], kfold_inputs.n_data_points)
    elpd_sum = combined_results["elpd_kfold"]
    se_elpd = combined_results["se_elpd_kfold"]
    p_sum = np.sum(kfold_results.ps)

    elpd_data = ELPDData(
        kind="loo_kfold",
        elpd=elpd_sum,
        se=se_elpd,
        p=p_sum,
        n_samples=kfold_inputs.n_samples,
        n_data_points=kfold_inputs.n_data_points,
        scale="log",
        warning=False,
        good_k=None,
        elpd_i=kfold_results.elpd_i if pointwise else None,
        pareto_k=None,
        n_folds=k,
    )

    if save_fits:
        elpd_data.fold_fits = kfold_results.fold_fits

    if pointwise:
        elpd_data.p_kfold_i = kfold_results.p_kfold_i

    return elpd_data



"""Compute exact Leave-One-Out cross validation refitting for problematic observations."""

from copy import deepcopy

import numpy as np
import xarray as xr
from arviz_base import rcParams
from xarray_einstats.stats import logsumexp

from arviz_stats.loo.helper_loo import _prepare_loo_inputs
from arviz_stats.loo.loo import loo
from arviz_stats.loo.wrapper import SamplingWrapper
from arviz_stats.utils import ELPDData

__all__ = ["reloo"]


def reloo(
    wrapper,
    loo_orig=None,
    var_name=None,
    log_weights=None,
    k_threshold=-np.inf,
    pointwise=None,
):
    r"""Recalculate exact Leave-One-Out cross validation refitting where the approximation fails.

    :func:`arviz_stats.loo` estimates the values of Leave-One-Out (LOO) cross validation using
    Pareto Smoothed Importance Sampling (PSIS) to approximate its value. PSIS works well when
    the posterior and the posterior_i (excluding observation i from the data used to fit)
    are similar. In some cases, there are highly influential observations for which PSIS
    cannot approximate the LOO-CV, and a warning of a large Pareto shape is sent by ArviZ.
    These cases typically have a handful of bad or very bad Pareto shapes, and a majority of
    good or ok shapes.

    Therefore, this may not indicate that the model is not robust enough nor that these
    observations are inherently bad, only that PSIS cannot approximate LOO-CV correctly.
    Thus, we can use PSIS for all observations where the Pareto shape is below a threshold
    and refit the model to perform exact cross validation for the handful of observations
    where PSIS cannot be used. This approach allows us to properly approximate LOO-CV with
    only a handful of refits, which in most cases is still much less computationally expensive
    than exact LOO-CV, which needs one refit per observation.

    Parameters
    ----------
    wrapper : SamplingWrapper
        An instance of a SamplingWrapper subclass that implements the necessary
        methods for model refitting. This wrapper allows ``reloo`` to work with
        any modeling framework.
    loo_orig : ELPDData, optional
        Existing LOO results with pointwise data. If None, will compute
        PSIS-LOO-CV first using the data from ``wrapper``.
    var_name : str, optional
        The name of the variable in log_likelihood groups storing the pointwise log
        likelihood data to use for loo computation. Defaults to None, which will
        use all log likelihood data.
    log_weights : DataArray or ELPDData, optional
        Smoothed log weights. Can be either:

        - A DataArray with the same shape as the log likelihood data
        - An ELPDData object from a previous :func:`arviz_stats.loo` call.

        Defaults to None. If not provided, it will be computed using the PSIS-LOO method.
    k_threshold : float, optional
        Pareto shape threshold. Observations with k values above this
        threshold will trigger a refit. Defaults to :math:`\min(1 - 1/\log_{10}(S), 0.7)`,
        where S is the number of samples.
    pointwise : bool, optional
        If True, return pointwise LOO data. Defaults to
        ``rcParams["stats.ic_pointwise"]``.

    Returns
    -------
    ELPDData
        Updated LOO results with the following attributes:

        - **kind**: "loo"
        - **elpd**: expected log pointwise predictive density
        - **se**: standard error of the elpd
        - **p**: effective number of parameters
        - **n_samples**: number of samples
        - **n_data_points**: number of data points
        - **scale**: "log"
        - **warning**: True if the estimated shape parameter of Pareto distribution is greater
          than ``good_k``.
        - **good_k**: For a sample size S, the threshold is computed as
          ``min(1 - 1/log10(S), 0.7)``
        - **elpd_i**: :class:`~xarray.DataArray` with the pointwise predictive accuracy, only if
          ``pointwise=True``
        - **pareto_k**: :class:`~xarray.DataArray` with Pareto shape values,
          only if ``pointwise=True``
        - **approx_posterior**: False (not used for standard LOO)
        - **log_weights**: Smoothed log weights.

    Notes
    -----
    It is strongly recommended to first compute :func:`arviz_stats.loo` on the inference results to
    confirm that the number of values above the threshold is small enough. Otherwise,
    prohibitive computation time may be needed to perform all required refits.

    As an extreme case, artificially assigning all ``pareto_k`` values to something
    larger than the threshold would make ``reloo`` perform the whole exact LOO-CV.
    This is not generally recommended nor intended, however, if needed, this function can
    be used to achieve the result.

    The term `exact cross-validation` is used here to refer to the process of refitting the model
    for each observation, which is also commonly called `brute-force cross-validation`. While
    this eliminates the PSIS approximation, other approximations remain in the process, such as
    those inherent in the fitting procedure itself and the use of cross-validation as a proxy
    for expected log predictive density (ELPD).

    Warnings
    --------
    Refitting can be computationally expensive. Check the number of high Pareto k
    values before using ``reloo`` to ensure the computation time is acceptable.

    See Also
    --------
    loo : Pareto smoothed importance sampling LOO-CV
    loo_moment_match : Moment matching for problematic observations

    References
    ----------

    .. [1] Vehtari et al. *Practical Bayesian model evaluation using leave-one-out cross-validation
        and WAIC*. Statistics and Computing. 27(5) (2017) https://doi.org/10.1007/s11222-016-9696-4
        arXiv preprint https://arxiv.org/abs/1507.04544.

    .. [2] Vehtari et al. *Pareto Smoothed Importance Sampling*.
        Journal of Machine Learning Research, 25(72) (2024) https://jmlr.org/papers/v25/19-556.html
        arXiv preprint https://arxiv.org/abs/1507.02646
    """
    if not isinstance(wrapper, SamplingWrapper):
        raise TypeError(
            "wrapper must be an instance of SamplingWrapper or a subclass. "
            "See the SamplingWrapper documentation for implementation details."
        )

    required_methods = ["sel_observations", "sample", "get_inference_data", "log_likelihood__i"]
    not_implemented = wrapper.check_implemented_methods(required_methods)

    if not_implemented:
        raise TypeError(
            "Passed wrapper instance does not implement all methods required for reloo "
            f"to work. Check the documentation of SamplingWrapper. {not_implemented} must be "
            "implemented and were not found."
        )

    pointwise = rcParams["stats.ic_pointwise"] if pointwise is None else pointwise

    if loo_orig is None:
        pareto_k = None
        if isinstance(log_weights, ELPDData):
            if log_weights.log_weights is None:
                raise ValueError("ELPDData object does not contain log_weights")
            pareto_k = log_weights.pareto_k
            log_weights = log_weights.log_weights

            # Dataset case comes from loo_subsample
            if isinstance(log_weights, xr.Dataset):
                if var_name and var_name in log_weights:
                    log_weights = log_weights[var_name]
                else:
                    log_weights = log_weights[list(log_weights.data_vars)[0]]

        loo_orig = loo(
            wrapper.idata_orig,
            var_name=var_name,
            pointwise=True,
            log_weights=log_weights,
            pareto_k=pareto_k,
        )

    if not isinstance(loo_orig, ELPDData):
        raise TypeError("loo_orig must be an ELPDData object.")

    if loo_orig.pareto_k is None or loo_orig.elpd_i is None:
        raise ValueError(
            "reloo requires pointwise LOO results with Pareto k values. "
            "Please compute the initial LOO with pointwise=True."
        )

    sample_dims = ["chain", "draw"]
    loo_refitted = deepcopy(loo_orig)

    loo_inputs = _prepare_loo_inputs(wrapper.idata_orig, var_name)
    obs_dims = loo_inputs.obs_dims
    n_data_points = loo_inputs.n_data_points
    n_samples = loo_inputs.n_samples
    log_likelihood = loo_inputs.log_likelihood

    if k_threshold is None:
        k_threshold = min(1 - 1 / np.log10(n_samples), 0.7)

    pareto_k_stacked = loo_orig.pareto_k.stack(__obs__=obs_dims).transpose("__obs__")
    bad_k_mask = pareto_k_stacked > k_threshold

    bad_obs_flat_indices = np.where(bad_k_mask.values)[0]

    if len(bad_obs_flat_indices) == 0:
        if not pointwise:
            loo_refitted.elpd_i = None
            loo_refitted.pareto_k = None
        return loo_refitted

    if not hasattr(loo_refitted, "p_loo_i") or loo_refitted.p_loo_i is None:
        loo_refitted.p_loo_i = xr.full_like(loo_refitted.elpd_i, np.nan)
        lpd_i = logsumexp(log_likelihood, b=1 / n_samples, dims=sample_dims)
        loo_refitted.p_loo_i = lpd_i - loo_refitted.elpd_i

    if len(obs_dims) == 1:
        bad_obs_indices = bad_obs_flat_indices
    else:
        obs_shape = [loo_orig.pareto_k.sizes[dim] for dim in obs_dims]
        bad_obs_indices = np.array(np.unravel_index(bad_obs_flat_indices, obs_shape)).T

    for i, _ in enumerate(bad_obs_flat_indices):
        if len(obs_dims) == 1:
            obs_idx = bad_obs_indices[i]
            coord_value = loo_orig.pareto_k.coords[obs_dims[0]].values[obs_idx]
            obs_idx_dict = {obs_dims[0]: coord_value}
            sel_idx = obs_idx
        else:
            obs_idx = bad_obs_indices[i]
            obs_idx_dict = {}
            for j, dim in enumerate(obs_dims):
                coord_value = loo_orig.pareto_k.coords[dim].values[obs_idx[j]]
                obs_idx_dict[dim] = coord_value
            sel_idx = obs_idx_dict

        log_lik_i = log_likelihood.sel(obs_idx_dict)
        hat_lpd_i = logsumexp(log_lik_i, dims=sample_dims, b=1 / n_samples).item()

        new_obs, excluded_obs = wrapper.sel_observations(sel_idx)
        fit = wrapper.sample(new_obs)
        idata_idx = wrapper.get_inference_data(fit)
        log_lik_idx = wrapper.log_likelihood__i(excluded_obs, idata_idx)
        elpd_loo_i = logsumexp(log_lik_idx, dims=sample_dims, b=1 / log_lik_idx.size).item()

        loo_refitted.elpd_i.loc[obs_idx_dict] = elpd_loo_i
        loo_refitted.pareto_k.loc[obs_idx_dict] = 0.0
        loo_refitted.p_loo_i.loc[obs_idx_dict] = hat_lpd_i - elpd_loo_i

    loo_refitted.elpd = np.sum(loo_refitted.elpd_i.values)
    loo_refitted.se = np.sqrt(n_data_points * np.var(loo_refitted.elpd_i.values))
    loo_refitted.p = np.sum(loo_refitted.p_loo_i.values)

    loo_refitted.warning = np.any(loo_refitted.pareto_k.values > loo_refitted.good_k)

    if not pointwise:
        loo_refitted.elpd_i = None
        loo_refitted.pareto_k = None
        loo_refitted.p_loo_i = None
        if hasattr(loo_orig, "log_weights") and loo_orig.log_weights is not None:
            loo_refitted.log_weights = loo_orig.log_weights
    else:
        if hasattr(loo_orig, "log_weights") and loo_orig.log_weights is not None:
            loo_refitted.log_weights = loo_orig.log_weights.copy()
            for i, obs_idx in enumerate(bad_obs_flat_indices):
                if len(obs_dims) == 1:
                    loo_refitted.log_weights[bad_obs_indices[i]] = np.nan
                else:
                    obs_idx_tuple = tuple(bad_obs_indices[i])
                    loo_refitted.log_weights[obs_idx_tuple] = np.nan

    return loo_refitted



"""Compute weighted expectations and predictive metrics using PSIS-LOO-CV."""

import numpy as np
import xarray as xr
from arviz_base import convert_to_datatree, extract
from xarray import apply_ufunc

from arviz_stats.loo.helper_loo import _warn_pareto_k
from arviz_stats.metrics import _metrics
from arviz_stats.utils import ELPDData, get_log_likelihood_dataset


def loo_expectations(
    data,
    var_name=None,
    log_weights=None,
    kind="mean",
    probs=None,
):
    """Compute weighted expectations using the PSIS-LOO-CV method.

    The expectations assume that the PSIS approximation is working well.
    The PSIS-LOO-CV method is described in [1]_ and [2]_.

    Parameters
    ----------
    data: DataTree or InferenceData
        It should contain the groups `posterior_predictive` and `log_likelihood`.
    var_name: str, optional
        The name of the variable in log_likelihood groups storing the pointwise log
        likelihood data to use for loo computation.
    log_weights : DataArray or ELPDData, optional
        Smoothed log weights. Can be either:

        - A DataArray with the same shape as the log likelihood data
        - An ELPDData object from a previous :func:`arviz_stats.loo` call.

        Defaults to None. If not provided, it will be computed using the PSIS-LOO method.
    kind: str, optional
        The kind of expectation to compute. Available options are:

        - 'mean': the mean of the posterior predictive distribution. Default.
        - 'median': the median of the posterior predictive distribution.
        - 'var': the variance of the posterior predictive distribution.
        - 'sd': the standard deviation of the posterior predictive distribution.
        - 'quantile': the quantile of the posterior predictive distribution.
    probs: float or list of float, optional
        The quantile(s) to compute when kind is 'quantile'.

    Returns
    -------
    loo_expec : DataArray
        The weighted expectations.
    khat : DataArray
        Function-specific Pareto k-hat diagnostics for each observation.

    Examples
    --------
    Calculate predictive 0.25 and 0.75 quantiles and the function-specific Pareto k-hat diagnostics

    .. ipython::

        In [1]: from arviz_stats import loo_expectations
           ...: from arviz_base import load_arviz_data
           ...: dt = load_arviz_data("radon")
           ...: loo_expec, khat = loo_expectations(dt, kind="quantile", probs=[0.25, 0.75])
           ...: loo_expec

    .. ipython::

        In [2]: khat

    References
    ----------

    .. [1] Vehtari et al. *Practical Bayesian model evaluation using leave-one-out cross-validation
        and WAIC*. Statistics and Computing. 27(5) (2017) https://doi.org/10.1007/s11222-016-9696-4
        arXiv preprint https://arxiv.org/abs/1507.04544.

    .. [2] Vehtari et al. *Pareto Smoothed Importance Sampling*.
        Journal of Machine Learning Research, 25(72) (2024) https://jmlr.org/papers/v25/19-556.html
        arXiv preprint https://arxiv.org/abs/1507.02646
    """
    if kind not in ["mean", "median", "var", "sd", "quantile"]:
        raise ValueError("kind must be either 'mean', 'median', 'var', 'sd' or 'quantile'")

    if kind == "quantile" and probs is None:
        raise ValueError("probs must be provided when kind is 'quantile'")

    dims = ("chain", "draw")
    if var_name is None:
        var_name = list(data.observed_data.data_vars.keys())[0]

    data = convert_to_datatree(data)
    log_likelihood = get_log_likelihood_dataset(data, var_names=var_name)
    n_samples = log_likelihood[var_name].sizes["chain"] * log_likelihood[var_name].sizes["draw"]

    if log_weights is None:
        log_weights, _ = log_likelihood.azstats.psislw()
        log_weights = log_weights[var_name]

    if isinstance(log_weights, ELPDData):
        if log_weights.log_weights is None:
            raise ValueError("ELPDData object does not contain log_weights")
        log_weights = log_weights.log_weights
        if var_name in log_weights:
            log_weights = log_weights[var_name]

    weights = np.exp(log_weights)

    posterior_predictive = extract(
        data, group="posterior_predictive", var_names=var_name, combined=False
    )

    weighted_predictions = posterior_predictive.weighted(weights)

    if kind == "mean":
        loo_expec = weighted_predictions.mean(dim=dims)

    elif kind == "median":
        loo_expec = weighted_predictions.quantile(0.5, dim=dims)

    elif kind == "var":
        # We use a Bessel's like correction term
        # instead of n/(n-1) we use ESS/(ESS-1)
        # where ESS/(ESS-1) = 1/(1-sum(weights**2))
        loo_expec = weighted_predictions.var(dim=dims) / (1 - np.sum(weights**2))
    elif kind == "sd":
        loo_expec = (weighted_predictions.var(dim=dims) / (1 - np.sum(weights**2))) ** 0.5
    else:  # kind == "quantile"
        loo_expec = weighted_predictions.quantile(probs, dim=dims)

    log_ratios = -log_likelihood[var_name]

    # Compute function-specific khat
    khat = apply_ufunc(
        _get_function_khat,
        posterior_predictive,
        log_ratios,
        input_core_dims=[dims, dims],
        output_core_dims=[[]],
        exclude_dims=set(dims),
        kwargs={"kind": kind},
        vectorize=True,
        dask="parallelized",
        output_dtypes=[float],
    )

    _warn_pareto_k(khat.values, n_samples)

    return loo_expec, khat


def loo_metrics(data, kind="rmse", var_name=None, log_weights=None, round_to="2g"):
    """Compute predictive metrics using the PSIS-LOO-CV method.

    Currently supported metrics are mean absolute error, mean squared error and
    root mean squared error. For classification problems, accuracy and balanced
    accuracy are also supported.

    The PSIS-LOO-CV method is described in [1]_ and [2]_.

    Parameters
    ----------
    data: DataTree or InferenceData
        It should contain groups `observed_data`, `posterior_predictive` and  `log_likelihood`.
    kind: str
        The kind of metric to compute. Available options are:

        - 'mae': mean absolute error.
        - 'mse': mean squared error.
        - 'rmse': root mean squared error. Default.
        - 'acc': classification accuracy.
        - 'acc_balanced': balanced classification accuracy.

    var_name: str, optional
        The name of the variable in log_likelihood groups storing the pointwise log
        likelihood data to use for loo computation.
    log_weights: DataArray or ELPDData, optional
        Smoothed log weights. Can be either:

        - A DataArray with the same shape as the log likelihood data
        - An ELPDData object from a previous :func:`arviz_stats.loo` call.

        Defaults to None. If not provided, it will be computed using the PSIS-LOO method.
    round_to: int or str, optional
        If integer, number of decimal places to round the result. If string of the
        form '2g' number of significant digits to round the result. Defaults to '2g'.

    Returns
    -------
    estimate: namedtuple
        A namedtuple with the mean of the computed metric and its standard error.

    Examples
    --------
    Calculate predictive root mean squared error

    .. ipython::

        In [1]: from arviz_stats import loo_metrics
           ...: from arviz_base import load_arviz_data
           ...: dt = load_arviz_data("radon")
           ...: loo_metrics(dt, kind="rmse")

    Calculate accuracy of a logistic regression model

    .. ipython::

        In [1]: dt = load_arviz_data("anes")
           ...: loo_metrics(dt, kind="acc")

    References
    ----------

    .. [1] Vehtari et al. *Practical Bayesian model evaluation using leave-one-out cross-validation
        and WAIC*. Statistics and Computing. 27(5) (2017) https://doi.org/10.1007/s11222-016-9696-4
        arXiv preprint https://arxiv.org/abs/1507.04544.

    .. [2] Vehtari et al. *Pareto Smoothed Importance Sampling*.
        Journal of Machine Learning Research, 25(72) (2024) https://jmlr.org/papers/v25/19-556.html
        arXiv preprint https://arxiv.org/abs/1507.02646
    """
    if var_name is None:
        var_name = list(data.observed_data.data_vars.keys())[0]

    observed = data.observed_data[var_name]
    predicted, _ = loo_expectations(data, kind="mean", var_name=var_name, log_weights=log_weights)

    return _metrics(observed, predicted, kind, round_to)


def _get_function_khat(
    values,
    log_weights,
    kind,
):
    """Compute function-specific k-hat diagnostics for LOO expectations.

    Parameters
    ----------
    values : ndarray
        Values of the posterior predictive distribution, raveled across sample dimensions.
    log_weights : ndarray
        Raw log weights from PSIS, raveled across sample dimensions.
    kind : str
        Type of expectation being computed ('mean', 'median', 'var', 'sd', 'quantile').

    Returns
    -------
    khat : float
        Function-specific k-hat estimate.
    """
    r_theta_da = xr.DataArray(
        np.exp(log_weights.ravel() - np.max(log_weights.ravel())), dims=["sample"]
    )

    # Get right tail khat
    try:
        khat_r_da = r_theta_da.azstats.pareto_khat(
            sample_dims="sample", tail="right", log_weights=False
        )
        khat_r = khat_r_da.item()
    except ValueError:
        khat_r = np.nan

    # For quantile/median, only need khat_r
    if kind in ["quantile", "median"]:
        return khat_r

    h_theta_values = values.ravel() if kind == "mean" else values.ravel() ** 2
    unique_h = np.unique(h_theta_values[np.isfinite(h_theta_values)])

    if len(unique_h) <= 1:
        return khat_r
    if len(unique_h) == 2:
        return khat_r
    if np.any(~np.isfinite(h_theta_values)):
        return khat_r

    hr_theta_da = xr.DataArray(h_theta_values * r_theta_da.values, dims=["sample"])

    try:
        khat_hr_da = hr_theta_da.azstats.pareto_khat(
            sample_dims="sample", tail="both", log_weights=False
        )
        khat_hr = khat_hr_da.item()
    except ValueError:
        khat_hr = np.nan

    if np.isnan(khat_hr) and np.isnan(khat_r):
        return np.nan
    return np.nanmax([khat_hr, khat_r])



"""Survival analysis utilities."""

import warnings

import numpy as np
from arviz_base import dict_to_dataset, extract


def kaplan_meier(
    dt,
    var_names,
    group="observed_data",
):
    """Compute Kaplan-Meier survival curves for observed data.

    Parameters
    ----------
    dt: DataTree
        DataTree with "posterior_predictive" and "observed_data" groups
    var_names : str or list of str
        The variables to compute the unique values.
    group : str
        The group from which to get the unique values.
    """
    if isinstance(var_names, str):
        var_names = [var_names]
    else:
        var_names = list(var_names)

    pp = extract(dt, group=group, var_names=var_names, keep_dataset=True, combined=False)
    try:
        constant_data = dt["constant_data"].dataset
    except KeyError:
        warnings.warn(
            "No 'constant_data' group found in DataTree. Assuming all events are observed."
        )
        constant_data = []

    dictio = {}

    for var_name in var_names:
        var = pp[var_name].values.flatten()
        if constant_data:
            if var_name in constant_data.data_vars:
                status = constant_data[var_name].values.flatten()
            else:
                warnings.warn(
                    f"No status variable found for {var_name} in 'constant_data'. "
                    "Assuming all events are observed."
                )
                status = np.ones_like(var)
        else:
            status = np.ones_like(var)
        sorted_indices = np.argsort(var)
        sorted_times = var[sorted_indices]
        sorted_status = status[sorted_indices]

        survival_probs = []
        survival_times = []
        current_prob = 1.0

        unique_times = np.unique(sorted_times)

        for t in unique_times:
            n_events = np.sum((sorted_times == t) & (sorted_status == 1))
            n_at_risk_t = np.sum(sorted_times >= t)

            if n_at_risk_t > 0 and n_events > 0:
                current_prob *= 1 - n_events / n_at_risk_t

            survival_times.append(t)
            survival_probs.append(current_prob)

        dictio[var_name] = np.stack([survival_times, survival_probs])

    return (
        dict_to_dataset(dictio)
        .rename({"draw": "km_points", "chain": "plot_axis"})
        .assign_coords({"plot_axis": ["x", "y"]})
    )


def generate_survival_curves(
    dt,
    var_names,
    group="posterior_predictive",
    num_samples=None,
    extrapolation_factor=1.2,
):
    """Compute Kaplan-Meier curves for predictive samples.

    Parameters
    ----------
    dt : DataTree
        DataTree with posterior_predictive and observed_data groups
    var_names : str or list of str
        The variables to compute the survival curves for.
    group : str, default "posterior_predictive"
        Group containing the predictive samples.
    num_samples : int, optional
        Number of samples to draw from the predictive distribution.
    extrapolation_factor : float, default 1.2
        Factor by which to limit the survival curves beyond the maximum observed time.
        Set to None to show the unaffected posterior predictive draws.
    """
    if isinstance(var_names, str):
        var_names = [var_names]
    else:
        var_names = list(var_names)
    # Extract predictive data
    pp = extract(dt, group=group, var_names=var_names, num_samples=num_samples, keep_dataset=True)
    obs_data = dt["observed_data"].dataset

    dictio = {}

    for var_name in pp.data_vars:
        # Handle extrapolation - get max observed time if needed
        max_observed_time = np.inf
        if extrapolation_factor is not None:
            if var_name in obs_data.data_vars:
                max_observed_time = obs_data[var_name].max().values

        # Find the maximum number of points across all samples first
        max_points = 0
        sample_data_list = []

        for i in range(num_samples):
            times = pp[var_name].isel(sample=i)

            # Filter times based on extrapolation factor
            if extrapolation_factor is not None:
                times = times[times <= max_observed_time * extrapolation_factor]

            # Skip if no times left after filtering
            if len(times) == 0:
                continue

            unique_times = np.unique(np.sort(times))
            max_points = max(max_points, len(unique_times))
            sample_data_list.append((i, times))

        # Now create aligned arrays
        n_valid_samples = len(sample_data_list)
        times_array = np.full((n_valid_samples, max_points), np.nan)
        probs_array = np.full((n_valid_samples, max_points), np.nan)

        for idx, (_, times) in enumerate(sample_data_list):
            # Create empirical survival function for this sample
            sorted_times = np.sort(times)
            unique_times = np.unique(sorted_times)

            for j, t in enumerate(unique_times):
                times_array[idx, j] = t
                probs_array[idx, j] = 1 - (j + 1) / len(unique_times)

        dictio[var_name] = np.stack([times_array, probs_array], axis=0)

    return (
        dict_to_dataset(dictio)
        .rename({"draw": "sample", "chain": "plot_axis"})
        .assign_coords({"plot_axis": ["x", "y"]})
    )



"""Continuously ranked probability scores with PSIS-LOO-CV weights."""

from collections import namedtuple

import numpy as np
import xarray as xr
from arviz_base import convert_to_datatree, extract
from xarray_einstats.stats import logsumexp

from arviz_stats.loo.helper_loo import (
    _get_r_eff,
    _prepare_loo_inputs,
    _validate_crps_input,
    _warn_pareto_k,
)
from arviz_stats.utils import round_num


def loo_score(
    data,
    var_name=None,
    log_weights=None,
    pareto_k=None,
    kind="crps",
    pointwise=False,
    round_to="2g",
):
    r"""Compute PWM-based CRPS/SCRPS with PSIS-LOO-CV weights.

    Implements the probability-weighted-moment (PWM) identity for the continuous ranked
    probability score (CRPS) with Pareto-smoothed importance sampling leave-one-out (PSIS-LOO-CV)
    weights, but returns its negative as a maximization score (larger is better). This assumes
    that the PSIS-LOO-CV approximation is working well.

    Specifically, the PWM identity used here is

    .. math::

        \operatorname{CRPS}_{\text{loo}}(F, y)
        = E_{\text{loo}}\left[|X - y|\right]
        + E_{\text{loo}}[X]
        - 2\cdot E_{\text{loo}} \left[X\,F_{\text{loo}}(X') \right].

    The PWM identity is described in [3]_, traditional CRPS and SCRPS are described in
    [1]_ and [2]_, and the PSIS-LOO-CV method is described in [4]_ and [5]_.

    Parameters
    ----------
    data : DataTree or InferenceData
        Input data. It should contain the ``posterior_predictive``, ``observed_data`` and
        ``log_likelihood`` groups.
    var_name : str, optional
        The name of the variable in the log_likelihood group to use. If None, the first
        variable in ``observed_data`` is used and assumed to match ``log_likelihood`` and
        ``posterior_predictive`` names.
    log_weights : DataArray, optional
        Smoothed log weights for PSIS-LOO-CV. Must have the same shape as the log-likelihood data.
        Defaults to None. If not provided, they will be computed via PSIS-LOO-CV. Must be provided
        together with ``pareto_k`` or both must be None.
    pareto_k : DataArray, optional
        Pareto tail indices corresponding to the PSIS smoothing. Same shape as the log-likelihood
        data. If not provided, they will be computed via PSIS-LOO-CV. Must be provided together with
        ``log_weights`` or both must be None.
    kind : str, default "crps"
        The kind of score to compute. Available options are:

        - 'crps': continuous ranked probability score. Default.
        - 'scrps': scale-invariant continuous ranked probability score.
    pointwise : bool, default False
        If True, include per-observation score values in the return object.
    round_to : int or str, default "2g"
        If integer, number of decimal places to round the result. If string of the form ``"2g"``,
        number of significant digits to round the result. Use None to return raw numbers.

    Returns
    -------
    namedtuple
        If ``pointwise`` is False (default), a namedtuple named ``CRPS`` or ``SCRPS`` with fields
        ``mean`` and ``se``. If ``pointwise`` is True, the namedtuple also includes a ``pointwise``
        field with per-observation values.

    Examples
    --------
    Compute scores and return the mean and standard error:

    .. ipython::
        :okwarning:

        In [1]: from arviz_stats import loo_score
           ...: from arviz_base import load_arviz_data
           ...: dt = load_arviz_data("centered_eight")
           ...: loo_score(dt, kind="crps")

    .. ipython::
        :okwarning:

        In [2]: loo_score(dt, kind="scrps")

    We can also pass previously computed PSIS-LOO weights and return the pointwise values:

    .. ipython::
        :okwarning:

        In [3]: from arviz_stats import loo
           ...: loo_data = loo(dt, pointwise=True)
           ...: loo_score(dt, kind="crps",
           ...:           log_weights=loo_data.log_weights,
           ...:           pareto_k=loo_data.pareto_k,
           ...:           pointwise=True)

    Notes
    -----
    For a single observation with posterior-predictive draws :math:`x_1, \ldots, x_S`
    and PSIS-LOO-CV weights :math:`w_i \propto \exp(\ell_i)` normalized so that
    :math:`\sum_{i=1}^S w_i = 1`, define the PSIS-LOO-CV expectation and the left-continuous
    weighted CDF as

    .. math::

        E_{\text{loo}}[g(X)] := \sum_{i=1}^S w_i\, g(x_i), \quad
        F_{\text{loo}}(x') := \sum_{i: x_i < x} w_i.

    The first probability-weighted moment is
    :math:`b_1 := E_{\text{loo}}\left[X\,F_{\text{loo}}(X')\right]`.
    With this, the nonnegative CRPS under PSIS-LOO-CV is

    .. math::

        \operatorname{CRPS}_{\text{loo}}(F, y)
        = E_{\text{loo}}\left[\,|X-y|\,\right]
        + E_{\text{loo}}[X] - 2\,b_1.

    For the scale term for the SCRPS, we use the PSIS-LOO-CV weighted Gini mean difference given by
    :math:`\Delta_{\text{loo}} := E_{\text{loo}}\left[\,|X - X'|\,\right]`.
    This admits the PWM representation given by

    .. math::

        \Delta_{\text{loo}} =
        2\,E_{\text{loo}}\left[\,X\,\left(2F_{\text{loo}}(X') - 1\right)\,\right].

    A finite-sample weighted order-statistic version of this is used in the function and is given by

    .. math::

        \Delta_{\text{loo}} =
        2 \sum_{i=1}^S w_{(i)}\, x_{(i)} \left\{\,2 F^-_{(i)} + w_{(i)} - 1\,\right\},

    where :math:`x_{(i)}` are the values sorted increasingly, :math:`w_{(i)}` are the
    corresponding normalized weights, and :math:`F^-_{(i)} = \sum_{j<i} w_{(j)}`.

    The locally scale-invariant score returned for ``kind="scrps"`` is

    .. math::

        S_{\text{SCRPS}}(F, y)
        = -\frac{E_{\text{loo}}\left[\,|X-y|\,\right]}{\Delta_{\text{loo}}}
        - \frac{1}{2}\log \Delta_{\text{loo}}.

    When PSIS weights are highly variable (large Pareto :math:`k`), Monte-Carlo noise can
    increase. This function surfaces PSIS-LOO-CV diagnostics via ``pareto_k`` and warns when
    tail behavior suggests unreliability.

    References
    ----------

    .. [1] Bolin, D., & Wallin, J. (2023). *Local scale invariance and robustness of
       proper scoring rules*. Statistical Science, 38(1), 140–159. https://doi.org/10.1214/22-STS864
       arXiv preprint https://arxiv.org/abs/1912.05642
    .. [2] Gneiting, T., & Raftery, A. E. (2007). *Strictly Proper Scoring Rules,
       Prediction, and Estimation*. Journal of the American Statistical Association,
       102(477), 359–378. https://doi.org/10.1198/016214506000001437
    .. [3] Taillardat M, Mestre O, Zamo M, Naveau P (2016). *Calibrated ensemble forecasts using
       quantile regression forests and ensemble model output statistics*. Mon Weather Rev
       144(6):2375–2393. https://doi.org/10.1175/MWR-D-15-0260.1
    .. [4] Vehtari, A., Gelman, A., & Gabry, J. (2017). *Practical Bayesian model
       evaluation using leave-one-out cross-validation and WAIC*. Statistics and Computing,
       27(5), 1413–1432. https://doi.org/10.1007/s11222-016-9696-4
       arXiv preprint https://arxiv.org/abs/1507.04544
    .. [5] Vehtari, A., et al. (2024). *Pareto Smoothed Importance Sampling*. Journal of
       Machine Learning Research, 25(72). https://jmlr.org/papers/v25/19-556.html
       arXiv preprint https://arxiv.org/abs/1507.02646
    """
    if kind not in {"crps", "scrps"}:
        raise ValueError(f"kind must be either 'crps' or 'scrps'. Got {kind}")

    data = convert_to_datatree(data)

    loo_inputs = _prepare_loo_inputs(data, var_name)
    var_name = loo_inputs.var_name
    log_likelihood = loo_inputs.log_likelihood

    y_pred = extract(data, group="posterior_predictive", var_names=var_name, combined=False)
    y_obs = extract(data, group="observed_data", var_names=var_name, combined=False)

    n_samples = loo_inputs.n_samples
    sample_dims = loo_inputs.sample_dims
    obs_dims = loo_inputs.obs_dims
    r_eff = _get_r_eff(data, n_samples)

    _validate_crps_input(y_pred, y_obs, log_likelihood, sample_dims=sample_dims, obs_dims=obs_dims)

    if (log_weights is None) != (pareto_k is None):
        raise ValueError(
            "Both log_weights and pareto_k must be provided together or both must be None. "
            "Only one was provided."
        )

    if log_weights is None and pareto_k is None:
        log_weights_da, pareto_k = log_likelihood.azstats.psislw(r_eff=r_eff, dim=sample_dims)
    else:
        log_weights_da = log_weights

    abs_error = np.abs(y_pred - y_obs)

    loo_weighted_abs_error = _loo_weighted_mean(abs_error, log_weights_da, sample_dims)
    loo_weighted_mean_prediction = _loo_weighted_mean(y_pred, log_weights_da, sample_dims)
    pwm_first_moment_b1 = _apply_pointwise_weighted_statistic(
        y_pred, log_weights_da, sample_dims, _compute_pwm_first_moment_b1
    )

    crps_pointwise = (
        loo_weighted_abs_error + loo_weighted_mean_prediction - 2.0 * pwm_first_moment_b1
    )

    if kind == "crps":
        pointwise_scores = -crps_pointwise
        khat_da = pareto_k
    else:
        gini_mean_difference = _apply_pointwise_weighted_statistic(
            y_pred, log_weights_da, sample_dims, _compute_weighted_gini_mean_difference
        )
        pointwise_scores = -(loo_weighted_abs_error / gini_mean_difference) - 0.5 * np.log(
            gini_mean_difference
        )
        khat_da = pareto_k

    _warn_pareto_k(khat_da, n_samples)

    n_pts = int(np.prod([pointwise_scores.sizes[d] for d in pointwise_scores.dims]))
    mean = pointwise_scores.mean().values.item()
    se = (pointwise_scores.std(ddof=0).values / (n_pts**0.5)).item()
    name = "SCRPS" if kind == "scrps" else "CRPS"

    if pointwise:
        return namedtuple(name, ["mean", "se", "pointwise"])(
            round_num(mean, round_to),
            round_num(se, round_to),
            pointwise_scores,
        )
    return namedtuple(name, ["mean", "se"])(
        round_num(mean, round_to),
        round_num(se, round_to),
    )


def _compute_pwm_first_moment_b1(values_sorted, weights):
    """Compute first PWM using a left-continuous weighted CDF."""
    values_sorted, weights_sorted = _sort_values_and_normalize_weights(values_sorted, weights)
    cumulative_weights = np.cumsum(weights_sorted)
    f_minus = cumulative_weights - weights_sorted
    return np.sum(weights_sorted * values_sorted * f_minus).item()


def _compute_weighted_gini_mean_difference(values, weights):
    """Compute PSIS-LOO-CV weighted Gini mean difference."""
    values_sorted, weights_sorted = _sort_values_and_normalize_weights(values, weights)
    cumulative_weights = np.cumsum(weights_sorted)
    cumulative_before = cumulative_weights - weights_sorted
    bracket = 2.0 * cumulative_before + weights_sorted - 1.0
    return (2.0 * np.sum(weights_sorted * values_sorted * bracket)).item()


def _loo_weighted_mean(values, log_weights, dim):
    """Compute PSIS-LOO-CV weighted mean."""
    log_num = logsumexp(log_weights, dims=dim, b=values)
    log_den = logsumexp(log_weights, dims=dim)
    return np.exp(log_num - log_den)


def _apply_pointwise_weighted_statistic(x, log_weights, sample_dims, stat_func):
    """Apply a weighted statistic over sample dims."""
    max_logw = log_weights.max(dim=sample_dims)
    weights = np.exp(log_weights - max_logw)
    stacked = "__sample__"
    xs = x.stack({stacked: sample_dims})
    ws = weights.stack({stacked: sample_dims})
    return xr.apply_ufunc(
        stat_func,
        xs,
        ws,
        input_core_dims=[[stacked], [stacked]],
        output_core_dims=[[]],
        vectorize=True,
        output_dtypes=[float],
    )


def _sort_values_and_normalize_weights(values, weights):
    """Sort values by ascending order and normalize weights."""
    idx = np.argsort(values, kind="mergesort")
    values_sorted = values[idx]
    weights_sorted = weights[idx]
    weights_sorted = weights_sorted / np.sum(weights_sorted)
    return values_sorted, weights_sorted



"""Functions for manipulating samples."""

import numpy as np
import xarray as xr
from arviz_base import convert_to_dataset

from arviz_stats.utils import get_array_function
from arviz_stats.validate import validate_dims


def thin(
    data,
    sample_dims="draw",
    group="posterior",
    var_names=None,
    filter_vars=None,
    coords=None,
    factor="auto",
    chain_axis=0,
    draw_axis=1,
):
    r"""Perform thinning.

    Thinning refers to retaining only every nth sample from a Markov Chain Monte Carlo (MCMC)
    simulation. This is usually done to reduce autocorrelation in the stored samples or simply
    to reduce the size of the stored samples.

    Parameters
    ----------
    data : array-like, DataArray, Dataset, DataTree, DataArrayGroupBy, DatasetGroupBy, or idata-like
        Input data. It will have different pre-processing applied to it depending on its type:

        - array-like: call array layer within ``arviz-stats``.
        - xarray object: apply dimension aware function to all relevant subsets
        - others: passed to :func:`arviz_base.convert_to_dataset`

    sample_dims : iterable of hashable, optional
        Dimensions to be considered sample dimensions and are to be reduced.
        Default ``rcParams["data.sample_dims"]``.
    group : hashable, default "posterior"
        Group on which to compute the ESS.
    var_names : str or list of str, optional
        Names of the variables for which the ess should be computed.
    filter_vars : {None, "like", "regex"}, default None
    coords : dict, optional
        Dictionary of dimension/index names to coordinate values defining a subset
        of the data for which to perform the computation.
    factor : str or int, default "auto"
        The thinning factor. If "auto", the thinning factor is computed based on bulk and tail
        effective sample size as suggested by Säilynoja et al. (2022) [1]_.
        If an integer, the thinning factor is set to that value.
    chain_axis, draw_axis : int, optional
        Integer indicators of the axis that correspond to the chain and the draw dimension.
        `chain_axis` can be ``None``.

    Returns
    -------
    ndarray, DataArray, Dataset, DataTree
        Thinned samples

    References
    ----------

    .. [1] Säilynoja, T., Bürkner, PC. & Vehtari, A. "Graphical test for discrete
           uniformity and its applications in goodness-of-fit evaluation and
           multiple sample comparison." Statistics and Computing 32(2), 32 (2022).
           https://doi.org/10.1007/s11222-022-10090-6

    Examples
    --------
    Thin the posterior samples using the default arguments:

    .. ipython::

        In [1]: from arviz_base import load_arviz_data
           ...: import arviz_stats as azs
           ...: data = load_arviz_data('non_centered_eight')
           ...: azs.thin(data)

    Thin a subset of the variables with a thinning factor of 10:

    .. ipython::

        In [1]: azs.thin(data, factor=10, var_names=["mu"])
    """
    if isinstance(data, list | tuple | np.ndarray):
        data = np.array(data)
        return get_array_function("thin")(
            data,
            factor=factor,
            chain_axis=chain_axis,
            draw_axis=draw_axis,
        )

    if isinstance(data, xr.core.groupby.DataArrayGroupBy | xr.core.groupby.DatasetGroupBy):
        # Make sure the grouped dimension is added as one of the dimensions to be reduced
        sample_dims = list(set(validate_dims(sample_dims)).union(data.group1d.dims))
        return data.map(
            thin,
            sample_dims=sample_dims,
            var_names=var_names,
            coords=coords,
            factor=factor,
        )

    if isinstance(data, xr.DataArray):
        if coords is not None:
            data = data.sel(coords)
        return data.azstats.thin(sample_dims=sample_dims, factor=factor)

    if isinstance(data, xr.DataTree):
        data = data.azstats.filter_vars(
            group=group, var_names=var_names, filter_vars=filter_vars
        ).datatree
        if coords is not None:
            data = data.sel(coords)
        return data.azstats.thin(sample_dims=sample_dims, group=group, factor=factor)

    data = convert_to_dataset(data, group=group)

    data = data.azstats.filter_vars(var_names=var_names, filter_vars=filter_vars).dataset
    if coords is not None:
        data = data.sel(coords)

    return data.azstats.thin(sample_dims=sample_dims, factor=factor)



"""Bayes Factor using Savage-Dickey density ratio."""

import warnings

from arviz_base import convert_to_datatree
from numpy import finfo
from scipy.interpolate import interp1d


def bayes_factor(data, var_names, ref_vals=0, return_ref_vals=False, prior=None):
    """
    Compute Bayes factor using Savage–Dickey ratio.

    Parameters
    ----------
    data : DataTree, or InferenceData
        The data object containing the posterior and optionally the prior distributions.
    var_names : str or list of str
        Names of the variables for which the bayes factor should be computed.
    ref_vals : float or list of float, default 0
        Reference value for each variable. Must match var_names in length if list.
    return_ref_vals : bool, default False
        If True, return the reference density values for the posterior and prior.
    prior : dict, optional
        Dictionary with prior distributions for each variable. If not provided,
        the prior will be taken from the `prior` group in the data object.

    Returns
    -------
    dict
        Dictionary with Bayes Factor values: BF10 and BF01 per variable.

    References
    ----------
    .. [1] Heck DW. *A caveat on the Savage-Dickey density ratio:
       The case of computing Bayes factors for regression parameters.*
       Br J Math Stat Psychol, 72. (2019) https://doi.org/10.1111/bmsp.12150

    Examples
    --------
    Compute Bayes factor for a home and intercept variable in a rugby dataset
    using a reference value of 0.15 for home and 3 for intercept.

    .. ipython::

        In [1]: from arviz_base import load_arviz_data
           ...: from arviz_stats import bayes_factor
           ...: dt = load_arviz_data("rugby")
           ...: bayes_factor(dt, var_names=["home", "intercept"], ref_vals=[0.15, 3])
    """
    data = convert_to_datatree(data)

    if isinstance(var_names, str):
        var_names = [var_names]

    if isinstance(ref_vals, int | float):
        ref_vals = [ref_vals] * len(var_names)

    if len(var_names) != len(ref_vals):
        raise ValueError("Length of var_names and ref_vals must match.")

    results = {}
    ref_density_vals = {}

    for var, ref_val in zip(var_names, ref_vals):
        posterior = data.posterior[var]
        prior = data.prior[var]

        if not isinstance(ref_val, int | float):
            raise ValueError(f"Reference value for variable '{var}' must be numerical")

        if ref_val > posterior.max() or ref_val < posterior.min():
            warnings.warn(
                f"Reference value {ref_val} for '{var}' is outside the posterior range. "
                "This may overstate evidence in favor of H1."
            )

        if ref_val > prior.max() or ref_val < prior.min():
            warnings.warn(
                f"Reference value {ref_val} for '{var}' is outside the prior range. "
                "Bayes factor computation is not reliable."
            )

        posterior_kde = posterior.azstats.kde(grid_len=512, circular=False)
        prior_kde = prior.azstats.kde(grid_len=512, circular=False)

        posterior_val = interp1d(
            posterior_kde.values[0],
            posterior_kde.values[1],
            bounds_error=False,
            fill_value=finfo("float").eps,
        )(ref_val).item()
        prior_val = interp1d(
            prior_kde.values[0],
            prior_kde.values[1],
            bounds_error=False,
            fill_value=finfo("float").eps,
        )(ref_val).item()

        if prior_val <= 0 or posterior_val <= 0:
            raise ValueError(
                f"Invalid KDE values at ref_val={ref_val}: "
                f"prior={prior_val}, posterior={posterior_val}"
            )

        bf_10 = prior_val / posterior_val
        bf_01 = 1 / bf_10
        results[var] = {"BF10": bf_10, "BF01": bf_01}

        if return_ref_vals:
            ref_density_vals[var] = {"prior": prior_val, "posterior": posterior_val}

    if return_ref_vals:
        return results, ref_density_vals

    return results



# pylint: disable=wildcard-import
"""Statistical computation and diagnostics for ArviZ."""

try:
    from arviz_stats.utils import *
    from arviz_stats.accessors import *
    from arviz_stats.loo import (
        loo,
        loo_i,
        loo_expectations,
        loo_metrics,
        loo_score,
        loo_pit,
        loo_approximate_posterior,
        loo_subsample,
        update_subsample,
        loo_moment_match,
        reloo,
        compare,
        SamplingWrapper,
        loo_kfold,
    )
    from arviz_stats.psense import psense, psense_summary
    from arviz_stats.metrics import kl_divergence, metrics, r2_score, wasserstein
    from arviz_stats.sampling_diagnostics import ess, mcse, rhat, rhat_nested
    from arviz_stats.summary import summary, ci_in_rope, mode
    from arviz_stats.manipulation import thin
    from arviz_stats.bayes_factor import bayes_factor
    from arviz_stats.visualization import ecdf, eti, hdi, histogram, kde, qds
    from arviz_stats.survival import kaplan_meier, generate_survival_curves

except ModuleNotFoundError:
    pass
